from docker.models.containers import Container
from docker.client import DockerClient, ContainerCollection

from typing import Any, List

from kubemon.dataclasses import Pair

import json

__all__ = ['get_container_pid', 'get_container_name']

def _is_container_or_str(container: Any) -> bool:
    return isinstance(container, Container) or isinstance(container, str)

def get_container_pid(container: Container) -> str:
    if not _is_container_or_str(container):
        raise TypeError('Argument must be of type docker.models.containers.Container or str')

    if isinstance(container, Container):
        path = f'/var/lib/docker/containers/{container.id}/config.v2.json'
    else:
        path = f'/var/lib/docker/containers/{container}/config.v2.json'

    with open(path, 'r') as fp:
        data = fp.read()
    data = json.loads(data)
    return data['State']['Pid']

def get_container_name(container: Container) -> str:
    if not _is_container_or_str(container):
        raise TypeError('Argument must be of type docker.models.containers.Container or str')

    if isinstance(container, Container):
        path = f'/var/lib/docker/containers/{container.id}/config.v2.json'
    else:
        path = f'/var/lib/docker/containers/{container}/config.v2.json'

    with open(path, 'r') as fp:
        data = fp.read()

    data = json.loads(data)

    name = data['Name']

    return name

def get_containers(client: DockerClient, namespace='', to_tuple=False) -> List[ContainerCollection]:
    """ 
    Returns a list of containers. 
        By default and for my research purpose I'm using Kubernetes, so I'm avoiding containers that
        contains 'POD' (created by Kuberenetes) and 'k8s-bigdata' (Namespace in Kubernetes that I've created) 
        in their name. Furthermore I assume that this filter is only applied when the platform is Linux-based, 
        because I've created a Kubernetes cluster only in Linux-based machines, otherwise, if the platform
        is Windows or MacOS, the function will return all containers that are running.

    Args:
        client (DockerClient): Object returned by docker.from_env()
        namespace (str): Used to filter containers created by Kubernetes. If empty, it returns all containers 
        except from 'kube-system' namespace
        to_tuple (bool): Return a list of namedtuples that represent a pair of container and container name 
    """
    containers = client.containers.list()
    _filter = lambda x: 'POD' not in x.name and namespace in x.name and 'kube-system' not in x.name

    if not to_tuple:
        return list(filter(_filter, containers))
    return list(map(lambda container: Pair(container, container.name), filter(_filter, containers)))



from kubemon.settings import Volatile
from kubemon.decorators import label
from ._docker import _load_metrics

import psutil

Volatile.set_procfs(psutil.__name__)

LABEL = 'cpu'

@label(LABEL)
def _os_cpu_usage(*args):
    metrics = {
        **psutil.cpu_stats()._asdict(),
        **psutil.cpu_times()._asdict(),
    }

    return metrics

@label(LABEL)
def _process_cpu_usage(pid: int):
    p = psutil.Process(pid)
    return p.cpu_times()._asdict()

@label(LABEL)
def _container_cpu_usage(*paths):
    metrics = dict()
    for path in paths:
        metrics.update(_load_metrics(path))
    return metrics


METRIC_FUNCTIONS = {
    'os': _os_cpu_usage,
    'process': _process_cpu_usage,
    'docker': _container_cpu_usage,
}


from kubemon.settings import Volatile

from pathlib import Path
from typing import List, Tuple
from requests import get as rget

import socket
import pickle
import fcntl
import struct
import psutil

Volatile.set_procfs(psutil.__name__)

__all__ = ['receive', 'send_to', 'get_default_nic', 'get_host_ip', 'is_alive', 'gethostname', 'get_json']

def receive(_socket: socket.socket, encoding_type='utf8', buffer_size=1024) -> str:
    """ 
    Wrapper function for receiving data from a socket. It also decodes it to utf8 by default.

    Args:
        _socket (socket): Socket that will be receiving data from;
        encoding_type (str): Encoding type for decoding incoming data.
    """
    # UDP
    addr = None
    if _socket.type == 2:
        # recv_size, _ = _socket.recvfrom(4)
        # buffer_size = struct.unpack('I', recv_size)[0]
        data, addr = _socket.recvfrom(buffer_size)
    else:
        # recv_size = _socket.recv(4)
        # buffer_size = struct.unpack('I', recv_size)[0]
        data = _socket.recv(buffer_size)

    data = pickle.loads(data, encoding=encoding_type)
    return data, addr

def send_to(_socket: socket.socket, data: object, address=tuple()) -> None:
    byte_data = pickle.dumps(data)
    # size = struct.pack('I', len(byte_data))

    # UDP
    if _socket.type == 2 and address:
        # _socket.sendto(size, address)
        _socket.sendto(byte_data, address)
    else:
        # _socket.send(size)
        _socket.send(byte_data)

def get_default_nic():
    """ Function to get the default network interface in a Linux-based system. """
    # Source: https://stackoverflow.com/a/20925510/12238188

    route = f'{psutil.PROCFS_PATH}/net/route'
    with open(route, mode='r') as fd:
        # Removing spaces and separating fields
        lines = list(map(lambda x: x.strip().split(), fd))
        
        # Removing header
        lines.pop(0)

        # Parsing nic
        for iface in lines:
            name = iface[0]
            dest = iface[1]
            flags = iface[3]

            if dest != '00000000' or not int(flags, 16) & 2:
                continue
            return name

def get_host_ip(ifname: str = "") -> str:
    """ Returns the actual IP address from the host """
    # Source: https://stackoverflow.com/a/24196955
    if not ifname:
        ifname = get_default_nic()

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(fcntl.ioctl(
        s.fileno(),
        0x8915,  # SIOCGIFADDR
        struct.pack('256s', ifname.encode())
    )[20:24])

def is_alive(address: str, port: int) -> bool:
    sockfd = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    
    try:
        sockfd.connect((address, port))
        ret = True
    except:
        ret = False
    
    sockfd.close()

    return ret

def gethostname() -> str:
    if Path('/etc/host_hostname').exists():
        with open('/etc/host_hostname', mode='r') as fp:
            return fp.read().strip()
    return socket.gethostname()

def nslookup(addr: str, port: int) -> List[str]:
    try:
        sockets = socket.getaddrinfo(addr, port)
        return [s[-1][0] for s in sockets if s[1] == socket.SOCK_STREAM]
    except socket.gaierror:
        return []

def get_json(url: str) -> Tuple[dict, int]:
    req = rget(url)

    if req.status_code == 200:
        return req.json(), req.status_code
    return dict(), req.status_code



from kubemon.log import create_logger
from kubemon.exceptions.platform_exception import NotLinuxException
from kubemon.cli import *
from kubemon.collector.commands import COMMAND_CLASSES

import kubemon.settings as settings

import sys

if __name__ == "__main__":

    if 'win' in sys.platform:
        raise NotLinuxException("Kubemon is only available for Linux-based Operating Systems. Sorry.")
        
    LOGGER = create_logger(__name__, level=settings.LOGGING_LEVEL)

    args = parser.parse_args()
    
    if not args.from_ctnr:
        settings.Volatile.set_procfs()
        LOGGER.info("PROCFS_PATH set to %s", '/proc')
    
    module = get_module(args.type, args.host, int(args.port))

    if module == None:
        LOGGER.info('Module %s does not exist', type(module))
        exit()

    if module != None and args.type == 'cli':
        module.run()

    if module != None and not args.type == 'cli':
        module.start()
  
    if args.list:
        print("Available modules:")
        LOGGER.debug("Listing modules")
        for module in MODULES:
            print(f"\t- {module.capitalize()}")
    
    if args.list_commands:
        print("Available commands:")
        LOGGER.debug("Listing commands")
        cmd = COMMAND_CLASSES.get('help')(COMMAND_CLASSES)
        print(cmd.execute())


from kubemon.metrics.disk import METRIC_FUNCTIONS
from kubemon.settings import Volatile
from .base_entity import BaseEntity

import psutil

Volatile.set_procfs(psutil.__name__)

class Disk(BaseEntity):
    """ Class for retrieving data about disk partitions from a Linux based system """

    def __init__(self, disk_name: str, *args, _disk_stat_path=f"{psutil.PROCFS_PATH}/diskstats", _partition_path=f"{psutil.PROCFS_PATH}/partitions"):
        self.__name = disk_name
        self._disk_stat_path = _disk_stat_path
        self._partition_path = _partition_path
        self._parse_device_driver()
        self.__sector_size = self._parse_sector_bytes(disk_name)
        self.__partitions = tuple(self._get_partitions(_partition_path))
        super(Disk, self).__init__(*args)

    @property
    def partitions(self):
        return self.__partitions

    @property
    def name(self):
        return self.__name

    @property
    def major(self):
        return self.__major
    
    @property
    def minor(self):
        return self.__minor
    
    @property
    def sector_size(self):
        return self.__sector_size
    
    def __str__(self):
        return f"Disk<maj={self.major}, min={self.minor}, name={self.name}>"
    
    def __repr__(self):
        return self.__str__()

    def get_usage(self, *args) -> dict:
        return METRIC_FUNCTIONS[self._monitor_type](*args)

    def _parse_sector_bytes(self, partition: str):
        """ Get disk sector size in bytes """
        with open(f"/sys/block/{partition}/queue/hw_sector_size", mode='r') as fd:
            return int(list(fd)[0])

    def _get_partitions(self, partition_path, prefix='sd'):
        """ Yields every disk partition that contains 'sd' or any given prefix """
        is_sd = lambda x: prefix in x
        with open(partition_path, mode='r') as f:
            lines = f.readlines()

        for line in lines:
            res = tuple(filter(is_sd, line.split()))
            if res:
                yield res[0]

    def _parse_device_driver(self):
        """ Parse device driver version """
        with open(self._disk_stat_path, mode='r') as fd:
            data = fd.readlines()
        
        # Split values from each line
        data = list(map(lambda x: x.split(), data))

        # Select only major, minor and name values
        data = list(map(lambda x: (int(x[0]), int(x[1]), x[2]), data))

        # Filter by device name
        data = list(filter(lambda x: x[2] == self.name, data))

        # Set values
        if data:
            self.__major = data[0][0]
            self.__minor = data[0][1]



from kubemon.utils.networking import send_to, is_alive
from kubemon.settings import DATA_PATH, MONITOR_PORT, START_MESSAGE, Volatile

from typing import Dict
from datetime import datetime
from functools import lru_cache, wraps

import time
import requests
import abc

def newline_decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        ret = func(*args, **kwargs)
        if not isinstance(ret, str):
            return str(ret) + '\n'
        return ret
    return wrapper


class Command(abc.ABC):
    def __init__(self, collector,):
        self._collector = collector
    
    @abc.abstractmethod
    def execute(self) -> str:
        pass

class StartCommand(Command):
    """ Start collecting metrics from all connected daemons in the collector.

    Args:
        - Directory name to be saving the data collected. Ex.: start test000
    """

    @newline_decorator
    def execute(self) -> str:
        collector = self._collector
        
        if not collector.dir_name:
            return "Argument 'dir_name' is missing."

        # Check if the number of daemons connected to collector 
        # are the number expected by Volatile.NUM_DAEMONS
        if Volatile.NUM_DAEMONS != len(collector):
            return f"{Volatile.NUM_DAEMONS} daemon(s) expected. There are connected only {len(collector)}."

        for sockfd in collector._monitor_sockets:
            send_to(sockfd, START_MESSAGE)

        with collector.mutex:
            collector.is_running = True

        collector.running_since = datetime.now()

        return f"Starting {len(collector)} daemons and saving data at {collector.address}:{str(DATA_PATH)}/{collector.dir_name}"


class InstancesCommand(Command):
    """ Lists all the connected monitor instances.
    """

    def _instances_per_daemon(self, module: str, url: str) -> str:
        req = requests.get(url + module)

        if req.status_code != 200:
            return ''
        return req.json()['total']

    @lru_cache
    @newline_decorator
    def execute(self) -> str:
        collector = self._collector
        message = ""
        port = MONITOR_PORT
        
        total = 0
        start_time = time.time()
        
        with collector.mutex:
            for hostname, addr in collector.daemons.items():
                url = f'http://{addr}:{port}'
                req = requests.get(url)

                if req.status_code != 200:
                    continue

                req = req.json()

                instances_per_module = {module: self._instances_per_daemon(module, url) for module in req['metric_paths']}
                total_per_daemon = sum(i for i in instances_per_module.values())
                total += req['total']
                message += f'Instances in {hostname}@{addr}: {total_per_daemon}\n'

                # Listing number of instances per module, per daemon
                for module, n_inst in instances_per_module.items():
                    name = module.upper().replace('/', '')
                    message += f'\t - {name}: {n_inst}\n'

        message += f'Total instances: {total}'  
        end_time = time.time()

        message += f'\nTotal processing time: {end_time - start_time:.3f}s'

        return message

class IsReadyCommand(Command):
    """ Returns the expected amount of instances that should 
    be connected to the collector.
    """

    @newline_decorator
    def execute(self) -> str:
        from kubemon.settings import Volatile
        
        collector = self._collector

        return not collector.stop_request and len(collector) == Volatile.NUM_DAEMONS

class ConnectedDaemonsCommand(Command):
    """ Lists all the daemons (hosts) connected.
    """
    
    @newline_decorator
    def execute(self) -> str:
        collector = self._collector

        daemons = collector.daemons

        message = f"Total daemons connected: {len(daemons)}\n"
        
        if len(daemons):
            for name, addr in daemons.items():
                message += "- " + name + " | " + addr + "\n"
        return message

class StopCommand(Command):
    """ Stop all monitors if they're running.
    """
    
    @newline_decorator
    def execute(self) -> str:
        collector = self._collector
        
        if not collector.is_running:
            return 'Collector isn\'t running to be stopped.'

        with collector.mutex:
            collector.stop_request = True
        
        # Release all the monitors from the barrier
        for _ in range(Volatile.NUM_DAEMONS):
            collector.barrier.acquire()

        # Resetting the collector status
        with collector.mutex:
            collector.stop_request = False
            collector.is_running = False

        return 'Stopped collector'

class NotExistCommand(Command):
    """ Indicates if a command does not exist. 
    """

    @newline_decorator
    def execute(self) -> str:
        return "Command does not exist"

class HelpCommand(Command):
    """ Lists all the available commands.
    """
    def __init__(self, command_dict: Dict[str, Command]):
        self._cmd_dict = command_dict
    
    @newline_decorator
    def execute(self) -> str:
        msg = ""

        for key, value in self._cmd_dict.items():
            # Ignoring dunder commands
            if not key.startswith('_'):
                msg += f'\'{key}\':'
                msg += value.__doc__ + '\n'
        
        return msg

class IsAliveCommand(Command):
    """ Tells if the collector is alive.
    """
    def __init__(self, address: str, port: int):
        self._addr = address
        self._port = port

    def execute(self) -> bool:
        return is_alive(self._addr, self._port)

COMMAND_CLASSES = {
    'start': StartCommand,
    'instances': InstancesCommand,
    'daemons': ConnectedDaemonsCommand,
    'stop': StopCommand,
    'not exist': NotExistCommand,
    'help': HelpCommand,
    'alive': IsAliveCommand,
    '_is_ready': IsReadyCommand,
}


from typing import Dict
from psutil import NoSuchProcess

from kubemon.decorators import label
from kubemon.settings import Volatile
from kubemon.utils.networking import get_default_nic
from kubemon.utils.process import pid_exists

import psutil
import re

Volatile.set_procfs(psutil.__name__)

LABEL = 'net'

def _to_digit(x: str) -> object:
    return int(x) if x.isdigit() else x

def _to_dict(cols, vals):
    return {k: v for k, v in zip(cols, vals)}

def parse_proc_net(pid: int) -> Dict:

    if not pid_exists(pid):
        raise NoSuchProcess(pid)

    fd = f"{psutil.PROCFS_PATH}/{pid}/net/dev"
    with open(fd, mode='r') as fd:
        lines = list(fd)

        # Removing trailing whitespaces
        lines = list(map(lambda x: re.split(r'\W+', x), lines))
        
        for i, line in enumerate(lines):
            # Removing empty strings
            line = list(filter(None, line))

            # Converting to int
            lines[i] = list(map(_to_digit, line))

        shift_factor = 9
        header = lines[1]
        
        # Labeling fields as tx or rx
        header[0] = 'iface'
        for i, field in enumerate(header[1:]):
            i = i + 1
            # shift_factor - 1 because its ignoring the 'face' field
            if i > shift_factor - 1:
                field = 'tx_' + field
            else:
                field = 'rx_' + field
            header[i] = field

        # Converting to dict
        for line in lines[2:]:
            yield _to_dict(header, line)

@label(LABEL)
def get_network_usage(pid: int) -> dict:
    """ Returns the network usage by a process 

        Args:
            pid (int): The pid of the process
    """

    if not pid_exists(pid):
        return NoSuchProcess(pid)

    nic = [nic for nic in parse_proc_net(pid) if nic['iface'] == get_default_nic()]
        
    if len(nic) == 0:
        ret = {
            'iface': 'any', 'rx_bytes': 0, 'rx_packets': 0, 'rx_errs': 0, 
            'rx_drop': 0, 'rx_fifo': 0, 'rx_frame': 0, 'rx_compressed': 0, 
            'rx_multicast': 0, 'tx_bytes': 0, 'tx_packets': 0, 'tx_errs': 0, 
            'tx_drop': 0, 'tx_fifo': 0, 'tx_colls': 0, 'tx_carrier': 0, 
            'tx_compressed': 0
        }
    else:
        ret = nic[0]

    ret.pop('iface')
    return ret

@label(LABEL)
def _get_os_net_usage(*args) -> dict:
    n = psutil.net_io_counters()
    nc = psutil.net_connections(kind='all')
    ret = {
        **n._asdict(),
        "num_connections": len(nc),
    }
    return ret

METRIC_FUNCTIONS = {
    'os': _get_os_net_usage,
    'process': get_network_usage,
    'docker': get_network_usage,
}


from docker.models.containers import Container
from kubemon.monitors.base import BaseMonitor

from kubemon.settings import DISK_PARTITION, Volatile
from kubemon.utils.networking import get_host_ip, gethostname
from kubemon.log import create_logger
from kubemon.entities import (
    CPU, Memory,
    Network, Disk
)

import psutil

LOGGER = create_logger(__name__)

class ProcessMonitor(BaseMonitor):
    __slots__ = ( 
        '__pid', '__container', 
        '__name', '__metrics'
    )

    _type = 'process'

    def __init__(self, container: Container, pid: int):
        Volatile.set_procfs(psutil.__name__)
        try:
            psutil.Process(pid)
        except psutil.NoSuchProcess:
            LOGGER.info('pid %s not exist', pid)
            exit(0)

        self.__container = container
        self.__pid = pid
        self.__metrics = {
            'cpu': CPU(self._type),
            'memory': Memory(self._type),
            'disk': Disk(DISK_PARTITION, self._type),
            'network': Network(self._type),
        }

    @property
    def pid(self) -> int:
        return self.__pid

    @property
    def container(self) -> Container:
        return self.__container

    def __str__(self) -> str:
        ip = get_host_ip().replace('.', '_')
        return f'ProcessMonitor_{gethostname()}_{ip}_{self.__container.name}_{self.__pid}'
    
    def __repr__(self) -> str:
        return f'<ProcessMonitor - {gethostname()} - {get_host_ip()} - {self.__container.name} - {self.__pid}>'

    def get_disk_usage(self) -> dict:
        """ Returns the disk usage by a process """

        return self.__metrics['disk'](self.pid)

    def get_memory_usage(self) -> dict:
        """ Returns the memory usage by a process """

        return self.__metrics['memory'](self.pid)

    def get_cpu_usage(self) -> dict:
        """ Returns the cpu usage by a process """

        return self.__metrics['cpu'](self.pid)

    def get_net_usage(self) -> dict:
        """ Returns the network usage by a process """
        
        return self.__metrics['network'](self.pid)

    def get_stats(self) -> dict:
        cpu = self.get_cpu_usage()
        io = self.get_disk_usage()
        mem = self.get_memory_usage()
        net = self.get_net_usage()

        return {**cpu, **io, **mem, **net}




from psutil import NoSuchProcess
from typing import Generator

from ._docker import _load_metrics
from kubemon.decorators import label
from kubemon.settings import Volatile
from kubemon.utils.process import pid_exists

import psutil

Volatile.set_procfs(psutil.__name__)

LABEL = 'disk'

class Partition:
    """ Class for a simple representation of a disk partition on a Linux based system.
        For more information about the attributes, please refer to https://www.kernel.org/doc/html/latest/block/stat.html
    """

    def __init__(self, partition_name, *fields, flush_io=0, flush_ticks=0):
        self.__partition_name = partition_name
        _fields = [
            'read_io', 'read_merge', 'read_sectors',
            'read_ticks', 'write_io', 'write_merge',
            'write_sectors', 'write_ticks', 'in_flight',
            'io_ticks', 'time_in_queue', 'discard_io',
            'discard_merges', 'discard_sectors','discard_ticks',
        ]

        self.__info = {k: v for k, v in zip(_fields, fields[:len(_fields)])}

        if flush_io and flush_ticks:
            self.__info.update({"flush_io": flush_io, "flush_ticks": flush_ticks})

    @property
    def name(self):
        return self.__partition_name

    def __getitem__(self, info):
        return self.__info[info]
    
    def items(self):
        return self.__info.items()

    def __str__(self):
        base = f'<Partition partition={self.name}, '
        metrics = ', '.join(f'{k}={v}' for k, v in self.__info.items())
        base += metrics
        base += '>'
        return base

    def __repr__(self):
        return self.__str__()

def _usage_generator(*partitions, all_partitions=False, partition='sda') -> Generator:
    """ Returns a generator for each partition """
    if not all_partitions:
        yield _parse_data(partition)
    else:
        for _partition in partitions:
            yield _parse_data(_partition)

def _parse_data(partition_name: str) -> Partition:
    """ Parses /sys/block/<partition_name>/stat file and returns a 'Partition' object"""
    filename = "/sys/block/{}/stat".format(partition_name)
    with open(filename, "r") as f:
        data = f.read()
    data = list(map(int, data.split()))
    return Partition(partition_name, *data)

@label(LABEL)
def _get_process_disk_usage(pid: int) -> dict:
    if not pid_exists(pid):
        raise NoSuchProcess(pid)
    
    p = psutil.Process(pid)

    return p.io_counters()._asdict()

@label(LABEL)
def _get_os_disk_usage(partition: str, all_partitions=False) -> dict:
    """ Get information from a partition or from all available partitions started with 'sd' """
    ret = tuple(_usage_generator(all_partitions, partition))

    if not all_partitions:
        return ret[0]
    return ret

def _rename_fields(fields: tuple) -> str:
    partition = fields[0]
    field = fields[1]
    value = fields[2]
    return f'{partition}_{field.lower()}', value
    
@label(LABEL)
def _get_container_disk_usage(*paths: str) -> dict:
    metrics = dict()
    for path in paths:
        data = _load_metrics(path)
        data = filter(lambda x: len(x) == 3, data)
        data = map(_rename_fields, data)
        data = {k: v for k, v in data}
        metrics.update(data)
    return metrics

METRIC_FUNCTIONS = {
    'os': _get_os_disk_usage,
    'process': _get_process_disk_usage,
    'docker': _get_container_disk_usage,
}


from functools import wraps, lru_cache
from typing import Any, Callable, Dict, List
from flask_restful import Resource, Api
from gunicorn.app.base import BaseApplication

from kubemon.dataclasses import Pod
from kubemon.log import create_logger
from kubemon.pod import list_pods
from kubemon.utils.process import get_children_pids, pid_exists
from kubemon.utils.containers import get_containers, get_container_pid
from kubemon.utils.networking import get_host_ip
from kubemon.utils.monitors import list_monitors

from kubemon.settings import (
    MONITOR_PORT,
    K8S_NAMESPACES,
    FROM_K8S,
    Volatile
) 
from . import (
    OSMonitor, 
    DockerMonitor, 
    ProcessMonitor
)

import requests
import psutil
import docker
import flask

__all__ = ['Kubemond']

LOGGER = create_logger('daemon')
APP = flask.Flask(__name__)
API = Api(APP)

def _get_stats(instances: list) -> Dict[str, dict]:
    return {str(i): i.get_stats() for i in instances}

class ROS(Resource):
    def get(self):
        instances = {
           'total': 1,
           'instances': _get_stats([OSMonitor()])
       }
        return flask.make_response(flask.jsonify(instances), 200)

class RDocker(Resource):
    def get(self):
        client = docker.from_env()
        instances = docker_instances(client)
        instances = {
           'total': len(instances),
           'instances': _get_stats(instances)
        }
        return flask.make_response(flask.jsonify(instances), 200)

class RProcess(Resource):
    def get(self):
       client = docker.from_env()
       instances = process_instances(client)
       instances = {
           'total': len(instances),
           'instances': _get_stats(instances)
       }
       return flask.make_response(flask.jsonify(instances), 200)

class RHome(Resource):
    def get(self):
        _MODULES = [DockerMonitor, ProcessMonitor, OSMonitor]
        data = {
            'hostname': str(OSMonitor()),
            'metric_paths': [flask.url_for(i) for i in list_monitors(_MODULES)],
            'total': sum(requests.get(flask.request.base_url + m).json()['total'] for m in list_monitors(_MODULES))
        }
        return flask.jsonify(data)

API.add_resource(RHome, '/')
API.add_resource(ROS, '/os', endpoint='os')
API.add_resource(RDocker, '/docker', endpoint='docker')
API.add_resource(RProcess, '/process', endpoint='process')

class Kubemond(BaseApplication):
    """ Kubemon Daemon Class. """

    def __init__(self, app=APP, port=MONITOR_PORT):
        Volatile.set_procfs(psutil.__name__)
        self.__port = port
        self.app = app
        self.logger = LOGGER
        self.options = {
                'workers': 4,
                'bind': f'{get_host_ip()}:{port}'
        }
        super().__init__()
    
    @property
    def port(self) -> int:
        return self.__port

    def load_config(self):
        config = {key: value for key, value in self.options.items()
                  if key in self.cfg.settings and value is not None}

        for k, v in config.items():
            self.cfg.set(k.lower(), v)

    def load(self):
        return self.app
    
    def start(self):
        self.run()

def client_error(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        if not len(args):
            raise ValueError('Missing argument \'client\'')

        client = args[0]
        if not isinstance(client, docker.client.DockerClient):
            raise TypeError(f'Must specify the DockerClient object instead of \'{type(client).__name__}\'')

        ret = func(*args, **kwargs)
        return ret
    return wrapper

@lru_cache
@client_error
def docker_instances(client: docker.client.DockerClient) -> List[DockerMonitor]:   
    to_monitor = lambda c: DockerMonitor(c.container) if isinstance(c, Pod) else DockerMonitor(c)

    pods = list_pods(*K8S_NAMESPACES, client=client, from_k8s=FROM_K8S)

    return list(map(to_monitor, pods))

@lru_cache
@client_error
def process_instances(client: docker.client.DockerClient) -> List[ProcessMonitor]:
    containers = get_containers(client)
    container_pids = [(get_container_pid(c), c) for c in containers]
    monitors = list()

    for pid, container_obj in container_pids:
        for cpid in get_children_pids(pid):
            if pid_exists(cpid):
                monitor = ProcessMonitor(container_obj, cpid)
                monitors.append(monitor)

    return monitors




from .entities.disk import Disk
from .entities.network import Network
from .entities.cpu import CPU
from .collector import Collector, CollectorClient
from .monitors import OSMonitor, DockerMonitor

__all__ = ['Disk', 'Network', 'CPU', 'Collector', 'OSMonitor', 'DockerMonitor', 'CollectorClient']


from typing import List

def parse_fields(data: List[List[str]]) -> dict:
    """
        Parse fields on cgroups files or any file that contains the following pattern:

        input -> [[str int],
                    [str int],
                    [str int],
                    ...]
        output -> {str: int, str: int, str: int, ...}
        
        Args:
            data (list): List of lists representing a pair of data
    """
    to_int = lambda x: int(x) if x.isdigit() else x
    data = list(map(lambda x: x.replace('\n', '').split(), data))
    field_count = list(filter(lambda x: len(x) > 2, data))

    # Checking columns to avoid dict exceptions
    if len(field_count) >= 1:
        ret = list(map(lambda x: tuple(map(to_int, x)), data))
    else:
        ret = {k: to_int(v) for k, v in data}

    return ret

def _load_metrics(path: str) -> list:
    try:
        with open(path, mode='r') as fd:
            return parse_fields(list(fd))
    except FileNotFoundError:
        return list()


from kubemon.metrics.network import METRIC_FUNCTIONS
from .base_entity import BaseEntity

class Network(BaseEntity):
    """ A simple object to return network usage """
    
    def get_usage(self, *args):
        return METRIC_FUNCTIONS[self._monitor_type](*args)



from kubemon.monitors.base import BaseMonitor
from kubemon.settings import DISK_PARTITION, Volatile
from kubemon.entities import (
    CPU, Network,
    Memory, Disk
)
from kubemon.utils.data import subtract_dicts, filter_dict
from kubemon.utils.containers import get_container_pid
from kubemon.utils.networking import gethostname, get_host_ip

from kubemon.log import create_logger
from kubemon.pod import *

from time import sleep
from typing import Callable, List
from docker.models.containers import Container
from pathlib import Path

import psutil
import functools
import logging

__all__ = ['DockerMonitor']

LOGGER = create_logger(__name__)

class StatParser:
    ''' Class to parse docker metrics to fit into a CSV format, by saving into a flat dictionary. '''

    def _parse_blkio_operations(data: List[dict]) -> dict:
        ''' Auxiliary function to StatParser.blkio to remove 'Major' and 'Minor' keys. 
            This function gets only the operations made by blkio counter.
        '''
        ret = dict()
        for d in data:
            ret[d['op'].lower()] = d['value']
        return ret
    
    @staticmethod
    def blkio(data: dict) -> dict:
        ''' Retrieves the blkio counter statistics from a container '''
        ret = dict()
        for key, value in data.items():
            if value:
                value = StatParser._parse_blkio_operations(value)
                for k, v in value.items():
                    ret[f'{key}_{k}'] = v
        return ret

    @staticmethod
    def cpu(data: dict) -> dict:
        ''' Retrieves the cpu counter statistics from a container '''
        name_cpu = lambda x, suffix: {f'cpu{i}_{suffix}': val for i, val in enumerate(x)}
        ret = dict()

        for key, value in data.items():
            if isinstance(value, dict):
                ret[key] = dict()
                for k, v in data[key].items():
                    ret[key][f'cpu_{k}'] = v
            else:
                ret[key] = value

        ret = {
            **name_cpu(data['cpu_usage']['percpu_usage'], 'usage'),
            'system_cpu_usage': data['system_cpu_usage'],
            **data['throttling_data']
        }
        return ret

    @staticmethod
    def memory(data: dict) -> dict:
        ''' Retrieves the memory counter statistics from a container '''
        fields = ['rss', 'cache', 'mapped_file', 'pgpgin', 'pgpgout', 'pgfault', 'pgmajfault', 'active_anon', 'inactive_anon', 'active_file', 'inactive_file', 'unevictable']
        ret = dict()

        for k, v in data.items():
            ret[f'memory_{k}'] = v 
        
        ret = filter_dict(ret, fields)
        return ret
    
    @staticmethod
    def network(data: dict) -> dict:
        ''' Retrieves the network counter statistics from a container '''
        ret = dict()

        # Here it flattens the 'networks' dictionary(ies)
        for net in data:
            for k, v in data[net].items():
                ret[f'net_{net}_{k}'] = v

        return ret

def log(logger: logging.Logger):
    def decorator(func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ret = func(*args, **kwargs)
            logger.info(f'Executed function {func.__name__}')
            return ret
        return wrapper
    return decorator

class DockerMonitor(BaseMonitor):
    __slots__ = ( 
        '__pid', '__container', 
        '__pod', '__pods', 
        '__name', '__stats_path',
        '_metrics', '_paths'
    )

    _type = 'docker'

    def __init__(self, container: Container, stats_path="/sys/fs/cgroup"):
        Volatile.set_procfs(psutil.__name__)
        
        self.__container = container
        self.__pid = get_container_pid(container)
        self.__stats_path = stats_path
        self._metrics = {
            'cpu': CPU(self._type),
            'network': Network(self._type),
            'disk': Disk(DISK_PARTITION, self._type),
            'memory': Memory(self._type)
        }

        self._paths = self._parse_cgroup(self.__pid, cgroup_path=stats_path)

    @property
    def stats_path(self):
        return self.__stats_path
    
    @property
    def pid(self) -> int:
        return self.__pid

    @property
    def container(self) -> Container:
        return self.__container

    def __str__(self) -> str:
        ip = get_host_ip().replace('.', '_')
        return f'DockerMonitor_{gethostname()}_{ip}_{self.__container.name}_{self.pid}'

    def __repr__(self) -> str:
        return f'<DockerMonitor - {gethostname()} - {get_host_ip()} - {self.__container.name} - {self.pid}>'

    @log(logger=LOGGER)
    def get_path(self, cgroup_controller: str, stat: str) -> str:
        """ 
        Get full path of the docker container within the pod, on cgroups directory 
        
        Args:
            cgroup_controller (str): cgroup controller. E.g. cpuacct, memory, blkio, etc.
            stat (str): file inside cgroup_controller
        """

        return self._paths[cgroup_controller] / stat

    def _parse_cgroup(self, container_pid: int, cgroup_path='/sys/fs/cgroup') -> dict:
        ignore = ['rdma', 'misc']

        path = f'{psutil.PROCFS_PATH}/{container_pid}/cgroup'

        with open(path, mode='r') as fp:
            data = fp.read()
        
        data = data.strip().split('\n')
        data = map(lambda x: x.split(':'), data)
        data = map(lambda x: x[1:], data)
        data = filter(lambda x: x[0] not in ignore, data)

        return {k: Path(cgroup_path + '/' + k + v) for k, v in data if k and not k.startswith('name')}

    @log(logger=LOGGER)
    def get_memory_usage(self) -> dict:
        """ Get the memory usage of a given container within a pod """
        path = self.get_path(cgroup_controller='memory', stat='memory.stat')

        return self._metrics['memory'](path)

    @log(logger=LOGGER)
    def get_disk_usage(self) -> dict:
        """ Get the disk usage of a given container within a pod """
        path = self.get_path(cgroup_controller='blkio', stat='blkio.throttle.io_service_bytes')
       
        return self._metrics['disk'](path)

    @log(logger=LOGGER)
    def get_cpu_times(self) -> dict:
        """ Get the CPU usage of a given container within a pod """
        path_cpuacct = self.get_path(cgroup_controller='cpu,cpuacct', stat='cpuacct.stat')
        path_cpu = self.get_path(cgroup_controller='cpu,cpuacct', stat='cpu.stat')

        return self._metrics['cpu'](path_cpuacct, path_cpu)

    @log(logger=LOGGER)
    def get_net_usage(self) -> dict:
        """ Get network usage of a given container within a pod 

        To understand why not gather information from cgroups, 
        please refer to https://docs.docker.com/config/containers/runmetrics/#network-metrics

        """

        return self._metrics['network'](self.pid)

    @log(logger=LOGGER)
    def get_stats(self) -> dict:
        """ Get all metrics of a given container within a pod """
        cpu = self.get_cpu_times()
        memory = self.get_memory_usage()
        network = self.get_net_usage()
        disk = self.get_disk_usage()

        return {**cpu, **memory, **network, **disk}
    


from kubemon.settings import DATA_PATH

from functools import reduce
from pathlib import Path
from typing import Any, List

from csv import DictWriter

__all__ = ['subtract_dicts', 'merge_dict', 'filter_dict', 'in_both']

def subtract_dicts(dict1: dict, dict2: dict, operation=lambda x, y: x-y) -> dict:
    """ Subtracts values from dict1 and dict2 """
    if len(dict1) != len(dict2):
        missing_keys = set(dict1.keys()) ^ set(dict2.keys())
        raise KeyError(f"Mapping keys {missing_keys} not found")
    values = map(lambda _dict: reduce(operation, _dict),
                 zip(dict2.values(), dict1.values()))
    return dict(zip(dict1.keys(), map(lambda n: round(n, 4), values)))


def merge_dict(*dicts: List[dict]) -> dict:
    """ Merges multiple dictionaries """
    if not len(dicts):
        raise ValueError('Must specify dictionaries')
    
    if not all(i for i in dicts if isinstance(i, dict)):
        raise TypeError('All arguments specified must be \'dict\'')

    ret = dict()
    for d in dicts:
        ret.update(d)
    return ret


def filter_dict(_dict: dict, *keys: List[Any]) -> dict:
    """ Apply a simple filter over a given dictionary
        Usage:
            >>> filter_dict({'a': 1, 'b': 2, 'c':3}, 'a', 'c')
            >>> {'a': 1, 'c': 3}
    """
    filters = keys
    if isinstance(keys[0], list):
        filters = keys[0]
    return {k: v for k, v in _dict.items() if k in filters}

def in_both(d1: dict, d2: dict) -> list:
    """ Return the intersecting keys between two dictionaries """
    if not isinstance(d1, dict) or not isinstance(d2, dict):
        d = d1 if not isinstance(d1, dict) else d2
        raise TypeError(f'Expected dictionary instead of {type(d).__name__}')
    
    if not d1 and not d2:
        return list()

    return list(set(d1.keys()).intersection(d2.keys()))

def save_csv(_dict: dict, name: str, dir_name="", output_dir=DATA_PATH) -> None:
    """ 
    Saves a dict into a csv 

    Args:
        _dict (dict): The dictionary that will be written or appended in the file
        name (str): The name of the file
        dir_name (str): Subdirectory inside .settings.DATA_PATH that the file will be saved

    Raises:
        TypeError 
            if `dir_name` type isn't string 
    """
    filename = f"{name}.csv"

    if not isinstance(_dict, dict):
        raise TypeError(f'Expected a dictionary instead of {type(_dict).__name__}')
    
    if not isinstance(name, str):
        raise TypeError(f'Expected a string instead of {type(_dict).__name__}')

    if dir_name and not isinstance(dir_name, str):
        raise TypeError(f"Expected str instead of {type(dir_name).__name__}")
    elif dir_name and isinstance(dir_name, str):
        output_dir = output_dir / dir_name

    Path(output_dir).mkdir(parents=True, exist_ok=True)

    output_dir = output_dir / filename

    write_header = False
    if not output_dir.exists():
        write_header = True

    with open(output_dir, mode='a+', newline="") as f:
        writer = DictWriter(f, _dict.keys())

        if write_header:
            writer.writeheader()

        writer.writerow(_dict)



from typing import List, Type

from kubemon.monitors.base import BaseMonitor

def list_monitors(type_list: List[Type]) -> List[str]:
    """ Returns the abbreviation of the monitor name """

    if not len(type_list):
        return list()

    if any(i for i in type_list if not issubclass(i, BaseMonitor)):
        not_subclass = [type(i).__name__ for i in type_list if not issubclass(i, BaseMonitor)]
        raise TypeError(f'Must specify a \'BaseMonitor\' inherited class instead of \'{not_subclass}\'')
    
    return [i._type for i in type_list]


from .docker_monitor import DockerMonitor
from .monitor import OSMonitor
from .process_monitor import ProcessMonitor
from .daemon import Kubemond

__all__ = ['DockerMonitor', 'OSMonitor', 'ProcessMonitor', 'Kubemond']


from threading import Thread
from logging import Logger
from datetime import datetime
from time import sleep
from os.path import join as join_path
from typing import Any, Dict, List

from kubemon.settings import START_MESSAGE, COLLECT_INTERVAL
from kubemon.utils.data import save_csv, subtract_dicts, in_both
from kubemon.utils.networking import receive, send_to, get_json


import socket

class MonitorHandler(Thread):
    def __init__(self, monitor_address: str, monitor_name: str, collector: object):
        self._addr = monitor_address
        self._monitor_name = monitor_name
        self._collector = collector
        self._logger = collector.logger

        self.paths = ""

        Thread.__init__(self)
    
    def __str__(self) -> str:
        return self._monitor_name

    @property
    def _log(self):
        if not self._logger:
            return
        
        if not isinstance(self._logger, Logger):
            raise TypeError('Must specify a logging.Logger object')
        
        return self._logger

    def _collect_per_path(self, paths: List[str]) -> Dict[str, Dict[str, Any]]:
        result = dict()
        for path in paths:
            json, _ = get_json(path)['instances']
            result.update(json)
        return result
    
    def _save(self, old: dict, new: dict, name: str) -> None:
        ret = subtract_dicts(old[name], new[name])
        ret['timestamp'] = datetime.now()

        dir_name = self._collector.dir_name
        if dir_name:
            dir_name = join_path(dir_name, name.split("_")[0])

        save_csv(ret, name, dir_name=dir_name)

    def _collect(self, paths: List[str]) -> None:
        old = self._collect_per_path(paths)

        sleep(COLLECT_INTERVAL)
        
        new = self._collect_per_path(paths)
        
        for monitor_name in in_both(old, new):
            self._save(old, new, monitor_name)
        
        self._log.info(f'Saved data of monitor {self}')

    def _check_paths(self, url: str) -> dict:
        if not self.paths:
            self.paths = self._collector._get_paths(url)

    def run(self) -> None:
        name = self._monitor_name
        addr = self._addr
        self._log.info(f"Starting MonitorHandler for {name}@{addr}")
        monitor_url = f'http://{self._addr}'
        
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sockfd:
            sockfd.connect((self._collector.address, self._collector.port))

            self._check_paths(monitor_url)

            send_to(sockfd, f'{name}@{addr}')

            while True:
                msg, _ = receive(sockfd)
                self._log.info('Received message: %s' % msg)

                if msg == START_MESSAGE:
                    self._log.info(f'Started collecting data from {name}@{addr}')

                    while not self._collector.stop_request:
                            
                        self._collect(self.paths)

                        if self._collector.stop_request:
                            self._log.info('Releasing the barrier.')
                            self._collector.barrier.release()
                        
                        if not self._collector.stop_request:
                            self._collector.loop_barrier.wait()
                            





from pathlib import Path

from dataclasses import dataclass

import logging
import sys
import os

def _check_environ(var: str) -> bool:
    return var in os.environ

VARS = (
    'MONITOR_PORT',
    'NUM_DAEMONS',
    'SERVICE_NAME',
    'COLLECT_INTERVAL',
    'OUTPUT_DIR',
    'NAMESPACES',
    'FROM_K8S',
)

LOGGING_LEVEL = logging.INFO
PROJECT_BASE = Path(__file__).absolute().parent.parent

if 'WITHIN_DOCKER' in os.environ:
    for var in VARS:
        if not _check_environ(var):
            raise EnvironmentError(f'Missing {var} env. var.')

    ## Monitors configuration
    MONITOR_PORT = int(os.environ['MONITOR_PORT'])
    COLLECT_INTERVAL = int(os.environ['COLLECT_INTERVAL'])
    NUM_DAEMONS = int(os.environ['NUM_DAEMONS'])
    SERVICE_NAME = os.environ['SERVICE_NAME']
    K8S_NAMESPACES = os.environ['NAMESPACES'].split()
    FROM_K8S = True if os.environ['FROM_K8S'].lower() == 'true' else False
    DATA_DIR = PROJECT_BASE / os.environ['OUTPUT_DIR']
else:
    MONITOR_PORT = 80
    COLLECT_INTERVAL = 5
    NUM_DAEMONS = 1
    SERVICE_NAME = 'monitor'
    K8S_NAMESPACES = ''
    FROM_K8S = False
    DATA_DIR = PROJECT_BASE / 'output'

## CLI configuration
CLI_PORT = 9880
COLLECTOR_HEALTH_CHECK_PORT = 9882

## Disk configuration
DISK_PARTITION = 'sda'

# Directories

## Logger path
LOG_PATH = DATA_DIR / 'logs'

## Data path
DATA_PATH = DATA_DIR / 'data'

# Starting message
START_MESSAGE = "OK"

# Creating base directory
if not DATA_DIR.exists():
    DATA_DIR.mkdir(parents=True)

# Creating log directory
if not LOG_PATH.exists():
    LOG_PATH.mkdir()

# Creating data directory
if not DATA_PATH.exists():
    DATA_PATH.mkdir()

nd = NUM_DAEMONS

@dataclass
class Volatile:
    PROCFS_PATH: str = '/procfs'
    NUM_DAEMONS: int = nd

    def set_procfs(module: str) -> None:
        sys.modules[module].PROCFS_PATH = Volatile.PROCFS_PATH
    
    def set_num_daemons(n: int) -> None:
        Volatile.NUM_DAEMONS = n



from typing import Callable, Any
from functools import wraps


def _label_metrics(preffix: str, metrics: dict) -> dict:
    return {f'{preffix}_{k}': v for k, v in metrics.items()}

def label(preffix: str) -> Callable:
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            ret = func(*args, **kwargs)
            return _label_metrics(preffix, ret)
        return wrapper
    return decorator



from kubemon.collector.commands import COMMAND_CLASSES, HelpCommand
from kubemon.collector.monitor_handler import MonitorHandler
from kubemon.log import create_logger
from kubemon.utils.networking import nslookup, receive, send_to, get_json
from kubemon.settings import ( 
    CLI_PORT,
    COLLECTOR_HEALTH_CHECK_PORT,
    MONITOR_PORT,
    SERVICE_NAME,
    Volatile
)

from typing import Dict, List
from time import sleep
from datetime import datetime

import socket
import os
import threading
import requests

def start_thread(func, args=tuple()):
    """
    Function to create and start a thread

        Args:
            func (function): Target function
            args (tuple): Target function arguments
    """
    threading.Thread(target=func, args=args).start()

LOGGER = create_logger(__name__)

class Collector(threading.Thread):
    def __init__(self, address: str, port: int, cli_port=CLI_PORT):
        threading.Thread.__init__(self)
        self.__address = address
        self.__port = port
        self.__cli_port = cli_port
        self.dir_name = None
        self.mutex = threading.Lock()
        self.__running_since = None
        self.__daemons = dict()
        self.barrier = threading.Semaphore(0)
        self.loop_barrier = threading.Barrier(Volatile.NUM_DAEMONS)
        self.stop_request = False
        self.is_running = False
        self.logger = LOGGER
        self.__metric_paths: List[str] = list()
        self._monitor_threads: List[str] = list()
        self._monitor_sockets: List[socket.socket] = list()
        self._alive = False

    @property
    def address(self) -> str:
        return self.__address

    @property
    def port(self) -> int:
        return self.__port
    
    @property
    def cli_port(self) -> int:
        return self.__cli_port
    
    @property
    def daemons(self) -> Dict[str, str]:
        return self.__daemons

    @property
    def running_since(self) -> str:
        return self.__running_since
    
    @running_since.setter
    def running_since(self, val: datetime) -> None:
        self.__running_since = val
 
    def __len__(self):
        return len(self._monitor_sockets)

    @property
    def metric_paths(self) -> List[str]:
        return self.__metric_paths

    def set_metric_paths(self, url: str) -> None:
        if not url.startswith('http://'):
            url = f'http://{url}'

        if not self.__metric_paths:
            resp = requests.get(url).json()

            with self.mutex:
                self.__metric_paths = resp['metric_paths']
                self.logger.info(f'Set metric paths to {self.metric_paths}')

    def _get_paths(self, url: str) -> str:
        self.set_metric_paths(url)

        return [url + p for p in self.metric_paths]

    def __listen_cli(self, cli: socket.socket) -> None:
        """ 
        Function to receive and redirect commands from a CLI to monitors. 
        Currently it is based on UDP sockets.

        Args:
            cli (socket.socket): client socket
        
        Returns:
            None
        """
        while True:
            self.logger.info(f'CLI waiting for connections')

            conn, addr = cli.accept()
            self.logger.info(f'CLI accepted connection with {addr[0]}:{addr[1]}')

            while True:
                self.logger.info('Waiting for commands')

                data, _ = receive(conn)

                # Splitting from 'command args' -> [command, args]
                data = data.split()

                self.logger.info(f"Received command '{data}' from {addr[0]}:{addr[1]}")

                message = ""
                if data:              
                    cmd = data[0].lower() # Command                      

                    command = COMMAND_CLASSES.get(cmd)

                    if not command:
                        command = COMMAND_CLASSES['not exist']

                    command = command(self)

                    if cmd == 'help':
                        command = HelpCommand(COMMAND_CLASSES)
                    elif cmd == 'start':
                        if len(data) >= 2:
                            self.dir_name = data[1]
                            self.logger.debug(f"dir_name set to {self.dir_name}")

                    message = command.execute()

                send_to(conn, message)

                self.logger.debug(f"Sending '{message}' to {addr[0]}:{addr[1]}")

                if cmd == 'exit':
                    break

    def setsock(self, address: str, port: int, socktype: socket.SocketKind) -> socket.socket:
        """ 
        Setup a server socket 
        
        Args:
            address (str): Address for binding
            port (int):    Port for binding
        """
        sockfd = socket.socket(socket.AF_INET, socktype)
        try:
            sockfd.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sockfd.bind((address, port))
        except Exception as e:
            self.logger.error(f"Error while trying to bind socket to port {port}. {e}")
            sockfd.close()
            exit(1)
        return sockfd

    def __start_cli(self) -> None:
        """ Wrapper function to setup CLI. """
        
        # Setup socket
        with self.setsock(self.address, self.cli_port, socket.SOCK_STREAM) as sockfd:
            self.logger.info(f'Started collector CLI at {self.address}:{self.cli_port}')
            sockfd.listen(3)

            # Start listening for commands
            self.__listen_cli(sockfd)

    def run(self) -> None:
        """ Start the collector """
        start_thread(self.__start_cli)
        start_thread(self.__is_alive)
        start_thread(self.__probe_for_daemons)

        self._listen_monitor_handler()

    def _check_new_services(self, new_svc: List[str]) -> List[str]:
        if not new_svc:
            return []

        current_daemons = list(self.__daemons.values())
        return set(new_svc) - set(current_daemons)

    def __probe_for_daemons(self) -> None:
        self._wait_alive()

        while True:
            sleep(1)
            services = nslookup(SERVICE_NAME, MONITOR_PORT)

            new_services = self._check_new_services(services)
            if not new_services:
                continue

            for service in new_services:
                url = f'http://{service}:{MONITOR_PORT}'

                json, status_code = get_json(url)

                if status_code != 200:
                    self.logger.info(f'Monitor {service} could not be registered.')
                    self.logger.info(f'Status code: {status_code}')
                    continue

                host_str = f'{json["hostname"]}@{service}:{MONITOR_PORT}'
                
                status = self._add_monitor(f'{service}:{MONITOR_PORT}', json['hostname'])

                if status:
                    self.logger.info(f'Registered monitor {host_str}')
                    
                    self.set_metric_paths(url)

                    with self.mutex:
                        self.__daemons[json['hostname']] = service



    def __is_alive(self) -> None:
        with self.setsock(self.address, COLLECTOR_HEALTH_CHECK_PORT, socket.SOCK_STREAM) as sockfd:
            sockfd.listen(2)

            while True:
                conn, _ = sockfd.accept()

                conn.close()

    def _wait_alive(self):
        while not self._alive:
            sleep(2)
        self.logger.info('I\'m alive')

    def _add_monitor(self, address: str, hostname: str) -> bool:
        if hostname not in self._monitor_threads:
            mh = MonitorHandler(address, hostname, self)

            mh.start()

            with self.mutex:
                self._monitor_threads.append(hostname)
            
            return True
        return False
    
    def _listen_monitor_handler(self) -> None:
        with self.setsock(self.address, self.port, socket.SOCK_STREAM) as sockfd:
            sockfd.listen()

            self._alive = True

            while True:
                conn, _ = sockfd.accept()
                
                hostname, _ = receive(conn)

                self.logger.info(f'Connection stablished between collector and {hostname}')

                self._monitor_sockets.append(conn)



import os
from kubemon.monitors.docker_monitor import DockerMonitor
from kubemon.monitors.monitor import OSMonitor
from kubemon.monitors.process_monitor import ProcessMonitor
from kubemon.utils.networking import _check_service
from kubemon.utils.data import diff_list

from kubemon.utils.monitors import list_monitors

import unittest
import os

class TestNetwork(unittest.TestCase):
    def test_check_service(self):
        service_name = 'MONITOR'
        self.assertFalse(_check_service(service_name))
    
    def test_check_service_exist(self):
        service_name = 'MONITOR'
        os.environ[service_name] = 'teste'

        self.assertTrue(_check_service(service_name))

class TestDaemon(unittest.TestCase):
    def test_modules(self):
        mods = list_monitors([DockerMonitor, OSMonitor, ProcessMonitor])

        self.assertEqual(mods, ['docker', 'os', 'process'])
    
    def test_modules_error(self):
        with self.assertRaises(TypeError):
            list_monitors([DockerMonitor, OSMonitor, str])
    
    def test_modules_empty(self):
        self.assertEqual(list_monitors([]), [])

class TestData(unittest.TestCase):
    def test_diff_lists(self):
        l1 = [1, 2, 3, 4]
        l2 = [1, 2, 3, 5]
        res = diff_list(l1, l2)

        self.assertEqual(res, [4, 5])
    
    def test_diff_lists_include(self):
        l1 = [1, 2, 3, 4]
        l2 = [1, 2, 3, 4, 5]
        res = diff_list(l1, l2)

        self.assertEqual(res, [5])

if __name__ == '__main__':
    unittest.main()


import abc
from typing import Dict

class BaseMonitor(abc.ABC):
    """ Abstract class for monitors. """

    @abc.abstractmethod
    def get_stats(self) -> Dict[str, dict]:
        """ Abstract method """


from kubemon.metrics.memory import METRIC_FUNCTIONS
from .base_entity import BaseEntity

class Memory(BaseEntity):
    """ Represents an Memory object that returns Memory usage of a system """

    def get_usage(self, *args):
        return METRIC_FUNCTIONS[self._monitor_type](*args)



from kubemon.exceptions.platform_exception import NotLinuxException

from abc import abstractmethod, ABC
from platform import platform

class BaseEntity(ABC):
    """ 
    Base class to represent a hardware component e.g. CPU, Hard Disk, RAM memory, network interfaces, etc. 
    
    Currently it only supports Linux based operating systems. Besides that, it raises an exception when
    inherited classes attepmts to run on a Windows OS. 
    """
    
    __platform = platform()

    def __init__(self, monitor_type: str):
        self._monitor_type = monitor_type
        self.__check_os()

    @abstractmethod
    def get_usage(self):
        """ Abstract method """
    
    def __call__(self, *args):
        return self.get_usage(*args)

    def __check_os(self):
        if not "Linux" in self.__platform:
            raise NotLinuxException(
                f"Using {self.__platform} instead of any Linux based system."
            )



from ..collector import *
from ..settings import MONITOR_PORT, COLLECT_INTERVAL
from ..monitors import Kubemond

import argparse

MODULES = ['collector', 'daemon']

__all__ = ['parser', 'get_modules']

parser = argparse.ArgumentParser(description='Kubemon commands')

parser.add_argument('-l', '--list', help="Lists all available modules", action='store_true')
parser.add_argument('-lc', '--list-commands', help='List all available commands for CollectorClient', action='store_true')
parser.add_argument('-t', '--type', help='Functionality of sys-monitor. E.g. collector, monitor, ...', required=True)
parser.add_argument('-H', '--host', help='Host that any of sys-monitor functions will be connecting', metavar='IP', required=True)
parser.add_argument('-p', '--port', default=MONITOR_PORT, help='Port of the host')
parser.add_argument('-c', '--command', default="", nargs='*', help="Command for be executing on CollectorClient")
parser.add_argument('-i', '--interval', default=COLLECT_INTERVAL, help="Data collection rate by monitors")
parser.add_argument(
    '-fc', '--from-container',
    dest='from_ctnr', action='store_false',
    default=True,
    help="Flag if the metrics are being collected from a container or from the host"
)
parser.add_argument('-n', '--num-daemons', default=1, help='Number of daemons that are expected to connect to the collector module.')

SYSTEMS = {
    'collector': Collector,
    'daemon': Kubemond,
    'cli': CollectorClient
}

def get_module(module: str, address: str, port: int) -> object:
    if module != 'daemon':
        return SYSTEMS.get(module)(address=address, port=port)
    return SYSTEMS.get(module)(port=port)


from dataclasses import dataclass

from docker.models.containers import Container

__all__ = ['Pair', 'Pod']

@dataclass
class Pair:
    """ Dataclass to represent a container Pair """
    container: Container
    name: str

@dataclass(frozen=True)
class Pod:
    """ Dataclass to represent a pod from kubernetes """
    namespace: str
    name: str
    container_name: str
    pod_id: str
    container_id: str
    container: Container



from typing import List

from kubemon.settings import Volatile

import psutil

Volatile.set_procfs(psutil.__name__)

__all__ = ['pid_exists', 'get_children_pids']

def pid_exists(pid: int) -> bool:
    return pid in psutil.pids()

def get_children_pids(pid: int) -> List[int]:
    with open(f'{psutil.PROCFS_PATH}/{pid}/task/{pid}/children', 'r') as fp:
        pids = list(fp)
    
    if len(pids): pids = pids[0]
    else:
        pids = ''
    
    return list(map(int, pids.split()))


import logging

from .settings import LOG_PATH, LOGGING_LEVEL

__all__ = ['create_logger']

def create_logger(name: str, level=LOGGING_LEVEL) -> logging.Logger:
    logger = logging.getLogger(name)

    logger.setLevel(level)

    formatter = logging.Formatter("[%(asctime)s] | %(levelname)s | [%(filename)s:%(lineno)s - %(funcName)10s()] | Logger %(name)s -> %(message)s")

    fh = logging.FileHandler(LOG_PATH.joinpath(f'{name}.log'), mode='a+')
    sh = logging.StreamHandler()
    
    fh.setLevel(level)
    sh.setLevel(level)

    fh.setFormatter(formatter)
    sh.setFormatter(formatter)

    logger.addHandler(fh)
    logger.addHandler(sh)

    return logger



from setuptools import setup, find_packages

VERSION = '2.2.1'
SHORT_DESCRIPTION = 'A tool for distributed container monitoring over Kubernetes'
REQUIRES = [
    'requests', 'psutil', 
    'docker', 'flask',
    'flask_restful',
    'gunicorn',
]

def _load_readme() -> str:
    with open('README.md', 'r') as fp:
        return fp.read()

setup(
    name='kubemon',
    version=VERSION,
    descripton=SHORT_DESCRIPTION,
    long_description=_load_readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/hrchlhck/kubemon',
    author='Pedro Horchulhack',
    author_email='pedro.horchulhack@ppgia.pucpr.br',
    license='MIT',
    packages=find_packages(),
    install_requires=REQUIRES
)



from kubemon.collector.commands import IsAliveCommand, COMMAND_CLASSES
from kubemon.utils.networking import receive, send_to, is_alive
from kubemon.settings import COLLECTOR_HEALTH_CHECK_PORT

from typing import Any
from time import sleep

import socket

__all__ = ["CollectorClient"]

def _check_alive(address: str, port: int) -> None:
    print('Waiting for collector to be alive')
    while not is_alive(address, port):
        sleep(1)
    else:
        print('Collector is alive!')

class ExecCommand:
    def __init__(self, address: str, port: int, cmd: str):
        self.__cmd = cmd
        self.__address = address
        self.__port = port

    @property
    def cmd(self) -> str:
        return self.__cmd
    
    @property
    def address(self):
        return self.__address
    
    @property
    def port(self):
        return self.__port
    
    def exec(self, *args) -> Any:
        if self.cmd == 'alive':
            data = IsAliveCommand(self.address, COLLECTOR_HEALTH_CHECK_PORT).execute()
            return data
        
        composed_cmd = self.cmd + " " + " ".join(map(str, args))
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sockfd:
            sockfd.connect((self.address, self.port))

            send_to(sockfd, composed_cmd)
            data, _ = receive(sockfd)
            print(data)

            send_to(sockfd, 'exit')
            _, _ = receive(sockfd)
            
            return data

class CollectorClient:
    def __init__(self, address: str, port: int):
        self.__address = address
        self.__port = port

        for k in COMMAND_CLASSES:
            setattr(self, k, ExecCommand(address, port, k).exec)
    
    @property
    def address(self):
        return self.__address
    
    @property
    def port(self):
        return self.__port

    def run(self) -> None:
        _check_alive(self.address, COLLECTOR_HEALTH_CHECK_PORT)

        sockfd = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sockfd.connect((self.address, self.port))

        try:
            cmd = ""
            while cmd != "exit":
                cmd = input(">>> ")
                
                if cmd == "alive":
                    data = IsAliveCommand(self.address, COLLECTOR_HEALTH_CHECK_PORT).execute()
                    print(data)
                    continue

                send_to(sockfd, cmd)

                data, _ = receive(sockfd)
                print(data)
        except KeyboardInterrupt:
            send_to(sockfd, 'exit')
            data, _ = receive(sockfd)
            print(data)
        finally:
            sockfd.close()



from docker.models.containers import Container
from typing import List
from collections import namedtuple

from kubemon.log import create_logger
from kubemon.dataclasses import Pod

import docker

__all__ = ['Pair', 'Pod']

Pair = namedtuple('Pair', ['name', 'id'])

LOGGER = create_logger(__name__)

def to_pod(container: Container) -> Pod:
    c_name = container.name
    fields = c_name.split('_')

    return Pod(fields[3], fields[2], fields[1], fields[4], container.id, container)

def list_pods(*namespaces, client=None, except_namespace='kube-system', from_k8s=True) -> List[Pod]:
    if client == None:
        client = docker.from_env()
    elif not isinstance(client, docker.client.DockerClient):
        raise TypeError(f'Must specify the DockerClient object instead of \'{type(client).__name__}\'')

    containers = client.containers.list()

    if not from_k8s:
        return containers

    if not any(c for c in containers if 'k8s' in c.name) or not len(namespaces):
        return list()

    containers = [c for c in containers if 'POD' not in c.name]

    # Get pods from a list of namespaces.
    # Otherwise, it returns all pods from all namespaces
    if not namespaces[0] == '*':
        containers = [c 
            for c in containers 
            for namespace in namespaces 
            if namespace in c.name \
            and except_namespace not in c.name
        ]

    return list(map(to_pod, containers))



from kubemon.utils.data import filter_dict
from kubemon.utils.process import pid_exists
from kubemon.settings import Volatile
from kubemon.decorators import label
from ._docker import _load_metrics

from psutil import NoSuchProcess
from typing import List

import psutil

Volatile.set_procfs(psutil.__name__)

LABEL = 'mem'

@label(LABEL)
def _get_system_memory_usage(*args):
    fields = [
        'nr_active_file', 'nr_inactive_file', 'nr_mapped', 
        'nr_active_anon', 'nr_inactive_anon', 'pgpgin', 
        'pgpgout', 'pgfree', 'pgfault', 'pgmajfault', 'pgreuse'
    ]
        
    def to_dict(nested_lists):
        atoms = map(lambda atom_list: atom_list.split(), nested_lists)
        ret = {k: int(v) for k, v in atoms}
        return ret

    with open(f"{psutil.PROCFS_PATH}/vmstat", mode="r") as fd:
        data = fd.readlines()

    ret = to_dict(data)

    return filter_dict(ret, fields)

@label(LABEL)
def _get_process_memory_usage(pid: int) -> dict:
    """ 
    Returns the memory usage based on /proc virtual file system available in the Linux kernel. 
    Any questions, please refer to https://man7.org/linux/man-pages/man5/proc.5.html

    Args:
        pid (int)
    """

    if not pid_exists(pid):
        raise NoSuchProcess(pid)

    fields = [
        'size', 'resident', 'shared',
        'text', 'lib', 'data', 'dt'
    ]

    with open(f'{psutil.PROCFS_PATH}/{pid}/statm', mode='r') as fd:
        ret = fd.read()

    return {k: int(v) for k, v in zip(fields, ret.split())}

@label(LABEL)
def _get_docker_memory_usage(*paths: List[str]) -> dict:
    fields = [
        'rss', 'cache', 'mapped_file', 'pgpgin', 'pgpgout', 
        'pgfault', 'pgmajfault', 'active_anon', 'inactive_anon', 
        'active_file', 'inactive_file', 'unevictable'
    ]

    metrics = dict()
    for path in paths:
        metrics.update(_load_metrics(path))
    
    return filter_dict(metrics, fields)

METRIC_FUNCTIONS = {
    'os': _get_system_memory_usage,
    'process': _get_process_memory_usage,
    'docker': _get_docker_memory_usage,
}


from kubemon.monitors.base import BaseMonitor
from kubemon.settings import DISK_PARTITION
from kubemon.entities import (
    CPU, Memory, 
    Network, Disk
)
from kubemon.utils.networking import get_host_ip, gethostname

class OSMonitor(BaseMonitor):
    _type = 'os'

    def get_stats(self):
        disk = Disk(DISK_PARTITION, self._type)
        cpu = CPU(self._type)
        net = Network(self._type)
        mem = Memory(self._type)

        return {**cpu(),**mem(),**disk(DISK_PARTITION),**net()}

    def __str__(self) -> str:
        ip = get_host_ip().replace('.', '_')
        return f'OSMonitor_{gethostname()}_{ip}'
    
    def __repr__(self) -> str:
        return f'<OSMonitor - {gethostname()} - {get_host_ip()}>'

